from typing import Any, List, Dict, Optional, Union, Generator, AsyncGenerator

from fastapi import Request

from orichain import error_explainer


class Generate(object):
    """
    Synchronous wrapper for AWSBedrock Anthropic's API client.

    This class provides methods for generating responses from AWSBedrock Anthropic's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize Anthropic AWSBedrock client and set up API keys.

        Args:
            - aws_access_key (str): access key
            - aws_secret_key (str): api key
            - aws_region (str): region name
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default is 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        if kwds.get("prompt_caching", True):
            self.prompt_caching = True
        else:
            self.prompt_caching = False

        from anthropic import AnthropicBedrock

        # Initialize the AWSBedrock Anthropic client with provided parameters
        self.client = AnthropicBedrock(
            aws_secret_key=kwds.get("aws_secret_key"),
            aws_access_key=kwds.get("aws_access_key"),
            aws_region=kwds.get("aws_region"),
            timeout=kwds.get(
                "timeout", Timeout(60.0, read=5.0, write=10.0, connect=2.0)
            ),
            max_retries=kwds.get("max_retries", 2),
        )

    def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: bool = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the AWSBedrock Anthropic model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message, chat_hist=chat_hist, do_json=do_json
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Setting system prompt in sampling params, as None type is not allowed
            if system_prompt:
                system = [{"type": "text", "text": system_prompt}]
                if self.prompt_caching:
                    system[0].update({"cache_control": {"type": "ephemeral"}})
                sampling_paras["system"] = system

            # Setting default max_tokens if not provided
            if "max_tokens" not in sampling_paras:
                sampling_paras["max_tokens"] = 512

            # Call the AWSBedrock Anthropic API with the formatted messages
            message = self.client.with_options(
                timeout=kwds.get("timeout")
            ).messages.create(
                messages=messages,
                model=model_name,
                **sampling_paras,
            )

            result = {
                "response": "{" + message.content[0].text
                if do_json
                else message.content[0].text,
                "metadata": {"usage": message.usage.to_dict()},
            }

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Generator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the AWSBedrock Anthropic model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            Generator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message, chat_hist=chat_hist, do_json=do_json
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages
            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Setting system prompt in sampling params, as None type is not allowed
                if system_prompt:
                    system = [{"type": "text", "text": system_prompt}]
                    if self.prompt_caching:
                        system[0].update({"cache_control": {"type": "ephemeral"}})
                    sampling_paras["system"] = system

                # Setting default max_tokens if not provided
                if "max_tokens" not in sampling_paras:
                    sampling_paras["max_tokens"] = 512

                # Start the streaming session
                with self.client.messages.stream(
                    messages=messages,
                    model=model_name,
                    **sampling_paras,
                ) as stream:
                    # Start JSON response if requested
                    if do_json:
                        yield "{"

                    # Stream text chunks as they become available
                    for chunk in stream.text_stream:
                        # Yield non-empty chunks
                        if chunk:
                            yield chunk

                # Get the final complete message after streaming
                final_response = stream.get_final_message()

                # Format the final response with metadata
                result = {
                    "response": "{" + final_response.content[0].text
                    if do_json
                    else final_response.content[0].text,
                    "metadata": {"usage": final_response.usage.to_dict()},
                }

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _chat_formatter(
        self,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        do_json: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the AWSBedrock Anthropic API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by AWSBedrock Anthropic's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message based on its type
            if isinstance(user_message, str):
                content = [{"type": "text", "text": user_message}]
                if self.prompt_caching:
                    content[0].update({"cache_control": {"type": "ephemeral"}})
                messages.append({"role": "user", "content": content})
            elif isinstance(user_message, List):
                messages.extend(user_message)
            else:
                raise KeyError(
                    """invalid user messages format, this error should not happen due to the validation steps being used.\
                    Please check the validation function"""
                )

            # Add JSON formatting instruction if requested
            if do_json:
                messages.append({"role": "assistant", "content": "{"})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncGenerate(object):
    """
    Asynchronous wrapper for AWSBedrock Anthropic's API client.

    This class provides methods for generating responses from AWSBedrock Anthropic's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize AWSBedrock Anthropic client and set up API keys.

        Args:
            - aws_access_key (str): access key
            - aws_secret_key (str): api key
            - aws_region (str): region name
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default is 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        if kwds.get("prompt_caching", True):
            self.prompt_caching = True
        else:
            self.prompt_caching = False

        from anthropic import AsyncAnthropicBedrock

        # Initialize the AWSBedrock Anthropic client with provided parameters
        self.client = AsyncAnthropicBedrock(
            aws_secret_key=kwds.get("aws_secret_key"),
            aws_access_key=kwds.get("aws_access_key"),
            aws_region=kwds.get("aws_region"),
            timeout=kwds.get(
                "timeout", Timeout(60.0, read=5.0, write=10.0, connect=2.0)
            ),
            max_retries=kwds.get("max_retries", 2),
        )

    async def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: bool = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the AWSBedrock Anthropic model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message, chat_hist=chat_hist, do_json=do_json
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Setting system prompt in sampling params, as None type is not allowed
            if system_prompt:
                system = [{"type": "text", "text": system_prompt}]
                if self.prompt_caching:
                    system[0].update({"cache_control": {"type": "ephemeral"}})
                sampling_paras["system"] = system

            # Setting default max_tokens if not provided
            if "max_tokens" not in sampling_paras:
                sampling_paras["max_tokens"] = 512

            # Check if the request was disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Call the AWSBedrock Anthropic API with the formatted messages
            message = await self.client.with_options(
                timeout=kwds.get("timeout")
            ).messages.create(
                messages=messages,
                model=model_name,
                **sampling_paras,
            )

            result = {
                "response": "{" + message.content[0].text
                if do_json
                else message.content[0].text,
                "metadata": {"usage": message.usage.to_dict()},
            }

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> AsyncGenerator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the AWSBedrock Anthropic model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message, chat_hist=chat_hist, do_json=do_json
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages
            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Setting system prompt in sampling params, as None type is not allowed
                if system_prompt:
                    system = [{"type": "text", "text": system_prompt}]
                    if self.prompt_caching:
                        system[0].update({"cache_control": {"type": "ephemeral"}})
                    sampling_paras["system"] = system

                # Setting default max_tokens if not provided
                if "max_tokens" not in sampling_paras:
                    sampling_paras["max_tokens"] = 512

                # Start the streaming session
                async with self.client.messages.stream(
                    messages=messages,
                    model=model_name,
                    **sampling_paras,
                ) as stream:
                    # Start JSON response if requested
                    if do_json:
                        yield "{"

                    # Stream text chunks as they become available
                    async for chunk in stream.text_stream:
                        # Check if the request was disconnected
                        if request and await request.is_disconnected():
                            yield {"error": 400, "reason": "request aborted by user"}
                            await stream.close()
                            break

                        # Yield non-empty chunks
                        if chunk:
                            yield chunk

                # Get the final complete message after streaming
                final_response = await stream.get_final_message()

                # Format the final response with metadata
                result = {
                    "response": "{" + final_response.content[0].text
                    if do_json
                    else final_response.content[0].text,
                    "metadata": {"usage": final_response.usage.to_dict()},
                }

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _chat_formatter(
        self,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        do_json: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the AWSBedrock Anthropic API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by AWSBedrock Anthropic's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message based on its type
            if isinstance(user_message, str):
                content = [{"type": "text", "text": user_message}]
                if self.prompt_caching:
                    content[0].update({"cache_control": {"type": "ephemeral"}})
                messages.append({"role": "user", "content": content})
            elif isinstance(user_message, List):
                messages.extend(user_message)
            else:
                raise KeyError(
                    """invalid user messages format, this error should not happen due to the validation steps being used.\
                    Please check the validation function"""
                )

            # Add JSON formatting instruction if requested
            if do_json:
                messages.append({"role": "assistant", "content": "{"})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Any, List, Dict, Optional, Union, Generator, AsyncGenerator
from botocore.eventstream import EventStream
import asyncio
from fastapi import Request
from orichain import error_explainer


class CreateAiter(object):
    """
    Asynchronous iterator wrapper to wrap synchronous iterator
    NOTE: This is currently only for the use of awsbedrock converse stream in async method
    """

    def __init__(self, event_stream: EventStream) -> None:
        """
        Convert EventStream(AWS) into a iterator
        """
        self.sync_streamer = iter(event_stream)
        self.SENTINEL = object()

    def __aiter__(self):
        return self

    async def __anext__(self):
        """
        Returns the next event stream asynchronously.
        """
        try:
            stream_output = next(self.sync_streamer)
        except StopIteration:
            stream_output = self.SENTINEL
        except Exception as e:
            error_explainer(e)
            stream_output = {"error": 500, "reason": str(e)}

        return stream_output


class Generate(object):
    """
    Synchronous wrapper for AWS Bedrock's API client.

    This class provides methods for generating responses from AWS Bedrock's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize AWSBedrock client and set up API keys.

        Args:
            - aws_access_key (str): access key
            - aws_secret_key (str): api key
            - aws_region (str): region name
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """

        from botocore.config import Config

        # Validate input parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("config") and not isinstance(kwds.get("config"), Config):
            raise TypeError(
                "Invalid 'config' type detected:",
                type(kwds.get("config")),
                ", Please enter valid config using:\n'from botocore.config import Config'",
            )
        else:
            pass

        if kwds.get("prompt_caching", True):
            self.prompt_caching = True
        else:
            self.prompt_caching = False

        import boto3

        # Initialize the AWSBedock boto client with provided parameters
        self.client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=kwds.get("aws_access_key"),
            aws_secret_access_key=kwds.get("aws_secret_key"),
            config=kwds.get("config")
            or Config(
                region_name=kwds.get("aws_region"),
                read_timeout=10,
                connect_timeout=2,
                retries={"total_max_attempts": 2},
                max_pool_connections=100,
            ),
        )

    def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the AWS Bedrock model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                do_json=do_json,
            )

            # Return early if message formatting failed
            if not isinstance(messages, List):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Setting up the request body
            body = {
                "modelId": model_name,
                "messages": messages,
                "inferenceConfig": sampling_paras,
                "additionalModelRequestFields": kwds.get("additional_model_fields", {}),
            }

            # Check for system_prompt
            if system_prompt:
                system = [{"text": system_prompt}]
                if self.prompt_caching:
                    system.append({"cachePoint": {"type": "default"}})
                body.update({"system": system})

            # Call the AWSBedrock client with the formatted messages
            result = self._generate_response(body=body)

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Generator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the AWS Bedrock model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            Generator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                do_json=do_json,
            )

            # Yield error and return early if message formatting failed
            if not isinstance(messages, List):
                yield messages
            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Setting up the request body
                body = {
                    "modelId": model_name,
                    "messages": messages,
                    "inferenceConfig": sampling_paras,
                    "additionalModelRequestFields": kwds.get(
                        "additional_model_fields", {}
                    ),
                }

                # Check for system_prompt
                if system_prompt:
                    system = [{"text": system_prompt}]
                    if self.prompt_caching:
                        system.append({"cachePoint": {"type": "default"}})
                    body.update({"system": system})

                # Start the streaming session
                streaming_response = self._stream_response(body=body)

                response = ""
                usage = None
                no_error = True

                # Stream text chunks as they become available
                for text in streaming_response:
                    if text and isinstance(text, str):
                        response += text
                        yield text
                    elif isinstance(text, Dict) and "error" not in text:
                        # Final chunk from AWS Bedrock while streaming
                        usage = text
                    elif isinstance(text, Dict) and "error" in text:
                        no_error = False
                        streaming_response.close()

                        # Yield non-empty chunks
                        yield text

                        break
                    else:
                        pass

                if no_error:
                    # Format the final response with metadata
                    result = {
                        "response": response.strip(),
                        "metadata": {"usage": usage},
                    }

                    yield result

        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _generate_response(self, body: Dict) -> Dict:
        """Converse function for generating response

        Args:
            body (Dict): Contains all the paras to pass

        Returns:
            Dict: Formatted response from the Converse"""
        try:
            # Call to Bedrock service from Converse method
            response = self.client.converse(**body)

            # Structuring response
            result = {
                "response": response.get("output", {})
                .get("message", {})
                .get("content", [{}])[0]
                .get("text", "")
                .strip(),
                "metadata": {"usage": response.get("usage", {})},
            }

            if response.get("metrics"):
                result["metadata"]["usage"].update(response.get("metrics"))

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def _stream_response(self, body: Dict) -> Generator:
        """ConverseStream function for generating response

        Args:
            body (Dict): Contains all the paras to pass

        Yeilds:
            Generator: Chunks of the model's response or error information"""
        try:
            # Call to Bedrock service from ConverseStream method
            response = self.client.converse_stream(**body)

            # Fetching generator
            streaming_response = response.get("stream")

            # Start the streaming session
            for event in streaming_response:
                # Waiting for text chunks to be generated
                if event.get("contentBlockDelta", {}).get("delta", {}).get("text"):
                    yield event.get("contentBlockDelta").get("delta").get("text")
                elif event.get("metadata", {}).get("usage"):
                    usage = event.get("metadata").get("usage")

                    if event.get("metadata").get("metrics"):
                        usage.update(event.get("metadata").get("metrics"))

                    yield usage

        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _chat_formatter(
        self,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        do_json: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the AWS Bedrock API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by AWS Bedrock's API

        Raises:
            KeyError: If the user message format is invalid

        NOTE: JSON METHOD UNSTABLE
        """
        try:
            messages = []

            # Add chat history if provided
            if chat_hist:
                for chat_log in chat_hist:
                    messages.append(
                        {
                            "role": chat_log.get("role"),
                            "content": [{"text": chat_log.get("content")}],
                        }
                    )

            # Add user message based on its type
            if isinstance(user_message, str):
                content = [
                    {
                        "text": user_message
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                        if do_json
                        else user_message
                    }
                ]
                if self.prompt_caching:
                    content.append({"cachePoint": {"type": "default"}})
                messages.append(
                    {
                        "role": "user",
                        "content": content,
                    }
                )
            elif isinstance(user_message, List):
                for logs in user_message:
                    messages.append(
                        {
                            "role": logs.get("role"),
                            "content": [{"text": logs.get("content")}],
                        }
                    )

                if messages[-1].get("role") == "user":
                    messages[-1]["content"][0]["text"] = (
                        messages[-1]["content"][0]["text"]
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                    )
                elif messages[-1].get("role") == "assistant":
                    messages[-2]["content"][0]["text"] = (
                        messages[-2]["content"][0]["text"]
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                    )
                else:
                    pass

            else:
                pass

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncGenerate(object):
    """
    Asynchronous wrapper for AWS Bedrock's API client.

    This class provides methods for generating responses from AWS Bedrock's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize AWSBedrock client and set up API keys.

        Args:
            - aws_access_key (str): access key
            - aws_secret_key (str): api key
            - aws_region (str): region name
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """

        from botocore.config import Config

        # Validate input parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("config") and not isinstance(kwds.get("config"), Config):
            raise TypeError(
                "Invalid 'config' type detected:",
                type(kwds.get("config")),
                ", Please enter valid config using:\n'from botocore.config import Config'",
            )
        else:
            pass

        if kwds.get("prompt_caching", True):
            self.prompt_caching = True
        else:
            self.prompt_caching = False

        import boto3

        # Initialize the AWSBedock boto client with provided parameters
        self.client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=kwds.get("aws_access_key"),
            aws_secret_access_key=kwds.get("aws_secret_key"),
            config=kwds.get("config")
            or Config(
                region_name=kwds.get("aws_region"),
                read_timeout=10,
                connect_timeout=2,
                retries={"total_max_attempts": 2},
                max_pool_connections=100,
            ),
        )

    async def __call__(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the AWS Bedrock model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                do_json=do_json,
            )

            # Return early if message formatting failed
            if not isinstance(messages, List):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Check if the request was disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Setting up the request body
            body = {
                "modelId": model_name,
                "messages": messages,
                "inferenceConfig": sampling_paras,
                "additionalModelRequestFields": kwds.get("additional_model_fields", {}),
            }

            # Check for system_prompt
            if system_prompt:
                system = [{"text": system_prompt}]
                if self.prompt_caching:
                    system.append({"cachePoint": {"type": "default"}})
                body.update({"system": system})

            # Call the AWSBedrock client with the formatted messages
            result = await self._generate_response(body=body)

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def streaming(
        self,
        model_name: str,
        user_message: Union[str, List[Dict[str, str]]],
        request: Optional[Request] = None,
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
        **kwds: Any,
    ) -> AsyncGenerator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the AWS Bedrock model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                do_json=do_json,
            )

            # Yield error and return early if message formatting failed
            if not isinstance(messages, List):
                yield messages
            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Setting up the request body
                body = {
                    "modelId": model_name,
                    "messages": messages,
                    "inferenceConfig": sampling_paras,
                    "additionalModelRequestFields": kwds.get(
                        "additional_model_fields", {}
                    ),
                }

                # Check for system_prompt
                if system_prompt:
                    system = [{"text": system_prompt}]
                    if self.prompt_caching:
                        system.append({"cachePoint": {"type": "default"}})
                    body.update({"system": system})

                # Start the streaming session
                streaming_response = self._stream_response(body=body)

                response = ""
                usage = None
                no_error = True

                # Stream text chunks as they become available
                async for text in streaming_response:
                    # Check if the request was disconnected
                    if request and await request.is_disconnected():
                        yield {"error": 400, "reason": "request aborted by user"}
                        await streaming_response.aclose()
                        break
                    elif text and isinstance(text, str):
                        response += text
                        yield text
                    elif isinstance(text, Dict) and "error" not in text:
                        # Final chunk from AWS Bedrock while streaming
                        usage = text
                    elif isinstance(text, Dict) and "error" in text:
                        no_error = False
                        await streaming_response.aclose()

                        # Yield non-empty chunks
                        yield text

                        break
                    else:
                        pass

                if no_error:
                    # Format the final response with metadata
                    result = {
                        "response": response.strip(),
                        "metadata": {"usage": usage},
                    }

                    yield result

        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _generate_response(self, body: Dict) -> Dict:
        """Converse function for generating response

        Args:
            body (Dict): Contains all the paras to pass

        Returns:
            Dict: Formatted response from the Converse"""
        try:
            # Call to Bedrock service from Converse method
            response = await asyncio.to_thread(self.client.converse, **body)

            # Structuring response
            result = {
                "response": response.get("output", {})
                .get("message", {})
                .get("content", [{}])[0]
                .get("text", "")
                .strip(),
                "metadata": {"usage": response.get("usage", {})},
            }

            if response.get("metrics"):
                result["metadata"]["usage"].update(response.get("metrics"))

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def _stream_response(self, body: Dict) -> AsyncGenerator:
        """ConverseStream function for generating response

        Args:
            body (Dict): Contains all the paras to pass

        Yeilds:
            AsyncGenerator: Chunks of the model's response or error information"""
        try:
            # Call to Bedrock service from ConverseStream method
            response = await asyncio.to_thread(self.client.converse_stream, **body)

            # Fetching generator
            streaming_response = CreateAiter(event_stream=response.get("stream"))

            # Use the async wrapper to iterate over events asynchronously.
            async for event in streaming_response:
                # Waiting for text chunks to be generated.
                if event is streaming_response.SENTINEL:
                    break
                elif event.get("contentBlockDelta", {}).get("delta", {}).get("text"):
                    text = event["contentBlockDelta"]["delta"].get("text", "")
                    yield text
                elif event.get("metadata", {}).get("usage"):
                    usage = event["metadata"]["usage"]
                    if event["metadata"].get("metrics"):
                        usage.update(event["metadata"].get("metrics"))
                    yield usage
                elif event.get("error"):
                    yield event
                else:
                    pass

        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _chat_formatter(
        self,
        user_message: Union[str, List[Dict[str, str]]],
        chat_hist: Optional[List[Dict[str, str]]] = None,
        do_json: Optional[bool] = False,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the AWS Bedrock API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by AWS Bedrock's API

        Raises:
            KeyError: If the user message format is invalid

        NOTE: JSON METHOD UNSTABLE
        """
        try:
            messages = []

            # Add chat history if provided
            if chat_hist:
                for chat_log in chat_hist:
                    messages.append(
                        {
                            "role": chat_log.get("role"),
                            "content": [{"text": chat_log.get("content")}],
                        }
                    )

            # Add user message based on its type
            if isinstance(user_message, str):
                content = [
                    {
                        "text": user_message
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                        if do_json
                        else user_message
                    }
                ]
                if self.prompt_caching:
                    content.append({"cachePoint": {"type": "default"}})
                messages.append(
                    {
                        "role": "user",
                        "content": content,
                    }
                )
            elif isinstance(user_message, List):
                for logs in user_message:
                    messages.append(
                        {
                            "role": logs.get("role"),
                            "content": [{"text": logs.get("content")}],
                        }
                    )

                if messages[-1].get("role") == "user":
                    messages[-1]["content"][0]["text"] = (
                        messages[-1]["content"][0]["text"]
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                    )
                elif messages[-1].get("role") == "assistant":
                    messages[-2]["content"][0]["text"] = (
                        messages[-2]["content"][0]["text"]
                        + "\n(Respond in JSON and do not give any explanation or notes)"
                    )
                else:
                    pass

            else:
                pass

            return messages

        except Exception as e:
            error_explainer(e)



from typing import Any, List, Union, Dict, Optional
import asyncio

from orichain import error_explainer


class DataBase(object):
    """
    Synchronous way to interact with the Pinecone knowledge base
    """

    def __init__(self, **kwds) -> None:
        """Initializes the Pinecone GRPC client.
        Args:
            api_key (str): Pinecone API key
            index_name (str): Pinecone index name
            namespace (str): Pinecone namespace
        Raises:
            KeyError: If required parameters are not found
        """
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("index_name"):
            raise KeyError("Required `index_name` not found")
        elif not kwds.get("namespace"):
            raise KeyError("Required `namespace` not found")

        from pinecone.grpc import PineconeGRPC

        self.namespace = kwds.get("namespace")
        client = PineconeGRPC(api_key=kwds.get("api_key"))
        self.index = client.Index(kwds.get("index_name"))

    def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves chunks from the knowledge base using query method
        Args:
            vector (List[float]): The query vector. This should be the same length as the dimension of the index
                                    being queried. Each `query()` request can contain only one of the parameters
                                    `id` or `vector`.. [optional]
            id (str): The unique ID of the vector to be used as a query vector.
                        Each `query()` request can contain only one of the parameters
                        `vector` or  `id`.. [optional]
            top_k (int): The number of results to return for each query. Must be an integer greater than 1.
            namespace (str): The namespace to fetch vectors from.
                                If not specified, the default namespace is used. [optional]
            filter (Dict[str, Union[str, float, int, bool, List, dict]]):
                    The filter to apply. You can use vector metadata to limit your search.
                    See https://www.pinecone.io/docs/metadata-filtering/.. [optional]
            include_values (bool): Indicates whether vector values are included in the response.
                                    If omitted the server will use the default value of False [optional]
            include_metadata (bool): Indicates whether metadata is included in the response as well as the ids.
                                        If omitted the server will use the default value of False  [optional]
            sparse_vector: (Union[SparseValues, Dict[str, Union[List[float], List[int]]]]): sparse values of the query vector.
                            Expected to be either a SparseValues object or a dict of the form:
                                {'indices': List[int], 'values': List[float]}, where the lists each have the same length.
        Returns:
            Dict: Result of retrieving the chunks
        Raises:
            ValueError: If `user_message_vector` or `id` is not provided
        """
        try:
            if not user_message_vector and not kwds.get("id"):
                raise ValueError(
                    "Atleast one is required `user_message_vector` or `id`"
                )

            # Querying the chunks from the knowledge base
            chunks = self.index.query(
                vector=user_message_vector,
                top_k=num_of_chunks,
                id=kwds.get("id"),
                sparse_vector=kwds.get("sparse_vector"),
                include_values=kwds.get("include_values"),
                filter=kwds.get("filter"),
                include_metadata=kwds.get("include_metadata", True),
                namespace=kwds.get("namespace") or self.namespace,
            )

            return chunks.to_dict()

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def fetch(self, ids: List[str], **kwds: Any) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base using fetch method
        Args:
            ids (List[str]): List of ids to fetch
            namespace (str, optional): The namespace to fetch vectors from.
                                If not specified, the default namespace is used.
        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = self.index.fetch(
                ids=ids,
                namespace=kwds.get("namespace") or self.namespace,
            )

            return chunks.to_dict()

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncDataBase(object):
    """
    Asynchronous way to interact with the Pinecone knowledge base
    """

    def __init__(self, **kwds) -> None:
        """Initializes the Pinecone GRPC client.
        Args:
            api_key (str): Pinecone API key
            index_name (str): Pinecone index name
            namespace (str): Pinecone namespace
        Raises:
            KeyError: If required parameters are not found
        """
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("index_name"):
            raise KeyError("Required `index_name` not found")
        elif not kwds.get("namespace"):
            raise KeyError("Required `namespace` not found")

        from pinecone.grpc import PineconeGRPC

        self.namespace = kwds.get("namespace")
        client = PineconeGRPC(api_key=kwds.get("api_key"))
        self.index = client.Index(kwds.get("index_name"))

    async def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves chunks from the knowledge base using query method
        Args:
            vector (List[float]): The query vector. This should be the same length as the dimension of the index
                                    being queried. Each `query()` request can contain only one of the parameters
                                    `id` or `vector`.. [optional]
            id (str): The unique ID of the vector to be used as a query vector.
                        Each `query()` request can contain only one of the parameters
                        `vector` or  `id`.. [optional]
            top_k (int): The number of results to return for each query. Must be an integer greater than 1.
            namespace (str): The namespace to fetch vectors from.
                                If not specified, the default namespace is used. [optional]
            filter (Dict[str, Union[str, float, int, bool, List, dict]]):
                    The filter to apply. You can use vector metadata to limit your search.
                    See https://www.pinecone.io/docs/metadata-filtering/.. [optional]
            include_values (bool): Indicates whether vector values are included in the response.
                                    If omitted the server will use the default value of False [optional]
            include_metadata (bool): Indicates whether metadata is included in the response as well as the ids.
                                        If omitted the server will use the default value of False  [optional]
            sparse_vector: (Union[SparseValues, Dict[str, Union[List[float], List[int]]]]): sparse values of the query vector.
                            Expected to be either a SparseValues object or a dict of the form:
                                {'indices': List[int], 'values': List[float]}, where the lists each have the same length.
        Returns:
            Dict: Result of retrieving the chunks
        Raises:
            ValueError: If `user_message_vector` or `id` is not provided
        """
        try:
            if not user_message_vector and not kwds.get("id"):
                raise ValueError(
                    "Atleast one is required `user_message_vector` or `id`"
                )

            # Querying the chunks from the knowledge base
            chunks = await asyncio.to_thread(
                self.index.query,
                vector=user_message_vector,
                top_k=num_of_chunks,
                id=kwds.get("id"),
                sparse_vector=kwds.get("sparse_vector"),
                include_values=kwds.get("include_values"),
                filter=kwds.get("filter"),
                include_metadata=kwds.get("include_metadata", True),
                namespace=kwds.get("namespace") or self.namespace,
            )

            return chunks.to_dict()

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def fetch(self, ids: List[str], **kwds: Any) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base using fetch method
        Args:
            ids (List[str]): List of ids to fetch
            namespace (str, optional): The namespace to fetch vectors from.
                                If not specified, the default namespace is used.
        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = await asyncio.to_thread(
                self.index.fetch,
                ids=ids,
                namespace=kwds.get("namespace") or self.namespace,
            )

            return chunks.to_dict()

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Dict, List, Optional, Generator, AsyncGenerator
from fastapi import Request
from orichain import error_explainer


class Generate(object):
    """
    Synchronous wrapper for Azure OpenAI's API client.

    This class provides methods for generating responses from OpenAI's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Azure OpenAI client and set up API key.

        Args:
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("azure_endpoint"):
            raise KeyError("Required `azure_endpoint` not found")
        elif not kwds.get("api_version"):
            raise KeyError("Required `api_version` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        self.model_name = kwds.get("model_name")

        from openai import AzureOpenAI
        import tiktoken

        # Initialize the Azure OpenAI client with provided parameters
        self.client = AzureOpenAI(
            azure_endpoint=kwds.get("azure_endpoint"),
            api_key=kwds.get("api_key"),
            api_version=kwds.get("api_version"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )

        self.tiktoken = tiktoken

    def __call__(
        self,
        model_name: str,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the Azure OpenAI model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Call the OpenAI API with the formatted messages
            completion = self.client.chat.completions.create(
                model=model_name,
                messages=messages,
                **sampling_paras,
                response_format={"type": "json_object"}
                if do_json
                else {"type": "text"},
            )

            result = {
                "response": completion.choices[0].message.content,
                "metadata": {"usage": completion.usage.to_dict()},
            }

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def streaming(
        self,
        model_name: str,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Generator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the Azure OpenAI model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            Generator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Start the streaming session
                completion = self.client.chat.completions.create(
                    model=model_name,
                    messages=messages,
                    stream=True,
                    **sampling_paras,
                    response_format={"type": "json_object"}
                    if do_json
                    else {"type": "text"},
                    stream_options={"include_usage": True},
                )

                response = ""

                # Stream text chunks as they become available
                for chunk in completion:
                    if chunk.choices and chunk.choices[0].delta.content:
                        response += chunk.choices[0].delta.content
                        yield chunk.choices[0].delta.content
                    elif chunk.usage:
                        usage = chunk.usage.to_dict()

                # Format the final response with metadata
                result = {
                    "response": response,
                    "metadata": {"usage": usage},
                }

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    def _chat_formatter(
        self,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the Azure OpenAI API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by Azure OpenAI's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add system prompt if provided
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message
            messages.append({"role": "user", "content": user_message})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def num_tokens_from_string(
        self, model_name: str = "gpt-3.5-turbo", string: str = None
    ) -> int:
        """Returns the number of tokens in a text string.

        Args:
        model_name (str): The tiktoken tokenizer for specifed model name
        string (str): String to calculate the tokens for

        Returns:
        int: Number of tokens"""
        if string and model_name:
            encoding = self.tiktoken.encoding_for_model(model_name)
            num_tokens = len(encoding.encode(string))
            return num_tokens
        else:
            return 0


class AsyncGenerate(object):
    """
    Asynchronous wrapper for Azure OpenAI's API client.

    This class provides methods for generating responses from Azure OpenAI's models
    both in streaming and non-streaming modes. It handles chat history formatting,
    error handling, and proper request configuration.
    """

    def __init__(self, **kwds) -> None:
        """
        Initialize Azure OpenAI client and set up API key.

        Args:
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        # Validate input parameters
        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif not kwds.get("azure_endpoint"):
            raise KeyError("Required `azure_endpoint` not found")
        elif not kwds.get("api_version"):
            raise KeyError("Required `api_version` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        self.model_name = kwds.get("model_name")

        from openai import AsyncAzureOpenAI
        import tiktoken

        # Initialize the Azure OpenAI client with provided parameters
        self.client = AsyncAzureOpenAI(
            azure_endpoint=kwds.get("azure_endpoint"),
            api_key=kwds.get("api_key"),
            api_version=kwds.get("api_version"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )

        self.tiktoken = tiktoken

    async def __call__(
        self,
        model_name: str,
        user_message: str,
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> Dict:
        """
        Generate a response from the specified model.

        Args:
            model_name (str): Name of the OpenAI model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False
            **kwds: Additional keyword arguments to pass to the client

        Returns:
            Dict: Response from the model or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Return early if message formatting failed
            if isinstance(messages, Dict):
                return messages

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}

            # Check if the request was disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Call the Azure OpenAI API with the formatted messages
            completion = await self.client.chat.completions.create(
                model=model_name,
                messages=messages,
                **sampling_paras,
                response_format={"type": "json_object"}
                if do_json
                else {"type": "text"},
            )

            result = {
                "response": completion.choices[0].message.content,
                "metadata": {"usage": completion.usage.to_dict()},
            }

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def streaming(
        self,
        model_name: str,
        user_message: str,
        request: Optional[Request] = None,
        chat_hist: Optional[List[str]] = None,
        sampling_paras: Optional[Dict] = None,
        system_prompt: Optional[str] = None,
        do_json: Optional[bool] = False,
    ) -> AsyncGenerator:
        """
        Stream responses from the specified model.

        Args:
            model_name (str): Name of the OpenAI model to use
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            request (Optional[Request], optional): FastAPI request object for connection tracking
            chat_hist (Optional[List[str]], optional): Previous conversation history
            sampling_paras (Optional[Dict], optional): Parameters for controlling the model's generation
            system_prompt (Optional[str], optional): System prompt to provide context to the model
            do_json (bool, optional): Whether to format the response as JSON. Defaults to False

        Yields:
            AsyncGenerator: Chunks of the model's response or error information
        """
        try:
            # Format the chat history and user message
            messages = await self._chat_formatter(
                user_message=user_message,
                chat_hist=chat_hist,
                system_prompt=system_prompt,
            )

            # Yield error and return early if message formatting failed
            if isinstance(messages, Dict):
                yield messages

            else:
                # Default empty dictionaries
                sampling_paras = sampling_paras or {}

                # Start the streaming session
                completion = await self.client.chat.completions.create(
                    model=model_name,
                    messages=messages,
                    stream=True,
                    **sampling_paras,
                    response_format={"type": "json_object"}
                    if do_json
                    else {"type": "text"},
                    stream_options={"include_usage": True},
                )

                response = ""

                # Stream text chunks as they become available
                async for chunk in completion:
                    if request and await request.is_disconnected():
                        yield {"error": 400, "reason": "request aborted by user"}
                        await completion.close()
                        break
                    else:
                        if chunk.choices and chunk.choices[0].delta.content:
                            response += chunk.choices[0].delta.content
                            yield chunk.choices[0].delta.content
                        elif chunk.usage:
                            usage = chunk.usage.to_dict()

                # Format the final response with metadata
                result = {
                    "response": response,
                    "metadata": {"usage": usage},
                }

                yield result
        except Exception as e:
            error_explainer(e)
            yield {"error": 500, "reason": str(e)}

    async def _chat_formatter(
        self,
        user_message: str,
        chat_hist: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        """
        Format user messages and chat history for the OpenAI API.

        Args:
            user_message (Union[str, List[Dict[str, str]]]): The user's message or formatted messages
            chat_hist (Optional[List[Dict[str, str]]], optional): Previous conversation history
            do_json (Optional[bool], optional): Whether to format the response as JSON. Defaults to False

        Returns:
            List[Dict]: Formatted messages in the structure expected by OpenAI's API

        Raises:
            KeyError: If the user message format is invalid
        """
        try:
            messages = []

            # Add system prompt if provided
            if system_prompt:
                messages.append({"role": "system", "content": system_prompt})

            # Add chat history if provided
            if chat_hist:
                messages.extend(chat_hist)

            # Add user message
            messages.append({"role": "user", "content": user_message})

            return messages

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def num_tokens_from_string(
        self, model_name: str = "gpt-3.5-turbo", string: str = None
    ) -> int:
        """Returns the number of tokens in a text string.

        Args:
        model_name (str): The tiktoken tokenizer for specifed model name
        string (str): String to calculate the tokens for

        Returns:
        int: Number of tokens"""
        if string and model_name:
            encoding = self.tiktoken.encoding_for_model(model_name)
            num_tokens = len(encoding.encode(string))
            return num_tokens
        else:
            return 0



from typing import Any, List, Dict, Union

from asyncio import gather
from concurrent.futures import ThreadPoolExecutor

from orichain import error_explainer


class Embed(object):
    """
    Synchronous Embed class to get embeddings from OpenAI API.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize OpenAI client and set up API key.

        Args:
            - api_key (str): OpenAI API key
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default is 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        from openai import OpenAI
        import tiktoken

        self.client = OpenAI(
            api_key=kwds.get("api_key"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )
        self.tiktoken = tiktoken

    def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            text (Union[str, List[str]]): Input text or list of texts
            model_name (str): Name of the embedding model to use
            **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            if isinstance(text, str):
                text = [text]

            with ThreadPoolExecutor() as executor:
                token_counts = list(
                    executor.map(
                        lambda sentence: self.num_tokens_from_string(
                            sentence, model_name
                        ),
                        text,
                    )
                )

            max_tokens = max(token_counts)

            if max_tokens > 8192:
                return {
                    "error": 400,
                    "reason": f"Embedding model maximum context length is 8192 tokens, \
                            however you requested {max_tokens} tokens. Please reduce text size",
                }

            response = self.client.embeddings.create(input=text, model=model_name)
            embeddings = [sentence.embedding for sentence in response.data]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def num_tokens_from_string(self, string: str, model_name: str) -> int:
        """Returns the number of tokens in a text string.
        Args:
            string (str): Input text
            model_name (str): Name of the embedding model
        Returns:
            (int): Number of tokens in the text string
        """
        encoding = self.tiktoken.encoding_for_model(model_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens


class AsyncEmbed(object):
    """
    Asynchronous Embed class to get embeddings from OpenAI API.
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize OpenAI client and set up API key.

        Args:
            - api_key (str): OpenAI API key
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default is 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default is 2

        Raises:
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter
        """
        from httpx import Timeout

        if not kwds.get("api_key"):
            raise KeyError("Required `api_key` not found")
        elif kwds.get("timeout") and not isinstance(kwds.get("timeout"), Timeout):
            raise TypeError(
                "Invalid 'timeout' type detected:",
                type(kwds.get("timeout")),
                ", Please enter valid timeout using:\n'from httpx import Timeout'",
            )
        elif kwds.get("max_retries") and not isinstance(kwds.get("max_retries"), int):
            raise TypeError(
                "Invalid 'max_retries' type detected:,",
                type(kwds.get("max_retries")),
                ", Please enter a value that is 'int'",
            )
        else:
            pass

        from openai import AsyncOpenAI
        import tiktoken

        self.client = AsyncOpenAI(
            api_key=kwds.get("api_key"),
            timeout=kwds.get("timeout")
            or Timeout(60.0, read=5.0, write=10.0, connect=2.0),
            max_retries=kwds.get("max_retries") or 2,
        )
        self.tiktoken = tiktoken

    async def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            text (Union[str, List[str]]): Input text or list of texts
            model_name (str): Name of the embedding model to use
            **kwargs: Additional keyword arguments for the embedding API

        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            if isinstance(text, str):
                text = [text]

            token_counts = await gather(
                *[
                    self.num_tokens_from_string(sentence, model_name)
                    for sentence in text
                ]
            )

            max_tokens = max(token_counts)

            if max_tokens > 8192:
                return {
                    "error": 400,
                    "reason": f"Embedding model maximum context length is 8192 tokens, \
                            however you requested {max_tokens} tokens. Please reduce text size",
                }

            response = await self.client.embeddings.create(input=text, model=model_name)
            embeddings = [sentence.embedding for sentence in response.data]

            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings
        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def num_tokens_from_string(self, string: str, model_name: str) -> int:
        """Returns the number of tokens in a text string.
        Args:
            string (str): Input text
            model_name (str): Name of the embedding model
        Returns:
            (int): Number of tokens in the text string
        """
        encoding = self.tiktoken.encoding_for_model(model_name)
        num_tokens = len(encoding.encode(string))
        return num_tokens



from typing import Any, List, Dict, Union
from orichain.embeddings import (
    openai_embeddings,
    awsbedrock_embeddings,
    stransformers_embeddings,
    azureopenai_embeddings,
)
import warnings
from orichain import hf_repo_exists

DEFUALT_EMBEDDING_MODEL = "text-embedding-ada-002"
MODEL_CLASS = {
    "text-embedding-ada-002": "OpenAI",
    "text-embedding-3-large": "OpenAI",
    "text-embedding-3-small": "OpenAI",
    "amazon.titan-embed-text-v1": "AWSBedrock",
    "amazon.titan-embed-text-v2:0": "AWSBedrock",
    "cohere.embed-english-v3": "AWSBedrock",
    "cohere.embed-multilingual-v3": "AWSBedrock",
}


class EmbeddingModel(object):
    """Synchronus Base class for embedding generation
    Default embedding model that will be used is `text-embedding-ada-002`"""

    default_model = DEFUALT_EMBEDDING_MODEL
    model_class = MODEL_CLASS

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Embedding Models class with the required parameters.
        Args:
            - model_name (str, optional): Name of the model to be used. Default: "gpt-4o-mini"

            ### Authentication parameters by provider:

            #### OpenAI models
            - api_key (str): OpenAI API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

            #### AWS Bedrock models
            - aws_access_key (str): AWS access key.
            - aws_secret_key (str): AWS secret key.
            - aws_region (str): AWS region name.
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

            #### Sentence Transformers models
            - model_download_path (str, optional): Path to download the model. Default: "/home/ubuntu/projects/models/embedding_models"
            - device (str, optional): Device to run the model. Default: "cpu"
            - trust_remote_code (bool, optional): Trust remote code. Default: False
            - token (str, optional): Hugging Face API token

            #### Azure OpenAI models
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - use_azure_openai (bool, optional): Use Azure OpenAI model. Default: False

        Raises:
            ValueError: If the model is not supported
            KeyError: If required parameters are missing
            TypeError: If the type of the parameter is incorrect
            ImportError: If the required library is not installed
        Warns:
            UserWarning: If no model_name is provided, defaulting to `text-embedding-ada-002`
        """

        # Check if the model name is provided
        if not kwds.get("model_name"):
            warnings.warn(
                f"\nNo 'model_name' specified, hence defaulting to {self.default_model}",
                UserWarning,
            )

        self.model_name = kwds.get("model_name", self.default_model)

        # Check if the model is supported
        if kwds.get("use_azure_openai"):
            self.model_type = "AzureOpenAI"
        elif hf_repo_exists(
            repo_id=self.model_name,
            repo_type=kwds.get("repo_type"),
            token=kwds.get("token"),
        ):
            self.model_type = "SentenceTransformer"
        else:
            self.model_type = self.model_class.get(self.model_name)

        if not self.model_type:
            raise ValueError(
                f"Unsupported model: {self.model_name}\n"
                f"Supported models are:\n"
                f"- "
                + "\n- ".join(list(self.model_class.keys()))
                + "\n- All sentence-transformers models"
            )

        # Initialize the model
        model_handler = {
            "OpenAI": openai_embeddings.Embed,
            "AWSBedrock": awsbedrock_embeddings.Embed,
            "SentenceTransformer": stransformers_embeddings.Embed,
            "AzureOpenAI": azureopenai_embeddings.Embed,
        }
        self.model = model_handler.get(self.model_type)(**kwds)

    def __call__(
        self, user_message: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """Get embeddings for the given text(s).
        Args:
            user_message (Union[str, List[str]]): Input text or list of texts
            **kwargs: Additional keyword arguments for the embedding API
        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        Raises:
            KeyError: If required parameters are missing
            TypeError: If the type of the parameter is incorrect
        Warns:
            UserWarning: If the model is not supported or if the model is not found in the model type
        """

        # Check if the model name is provided
        if kwds.get("model_name"):
            # Check if the model is supported in the model type class
            if self.model_class.get(kwds.get("model_name")) == self.model_type:
                model_name = kwds.get("model_name")
            elif self.model_type == "SentenceTransformer":
                warnings.warn(
                    f"\nFor using different sentence-transformers model: {kwds.get('model_name')}\n"
                    f"again reinitialize the EmbeddingModels class as currently {self.model_name} is already loaded"
                    f"Hence defaulting the model to {self.model_name}",
                    UserWarning,
                )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
            else:
                warnings.warn(
                    f"Unsupported model: {kwds.get('model_name')}\nSupported models are:"
                    f"\n- "
                    + "\n- ".join(list(self.model_class.keys()))
                    + "\n- All sentence-transformers models\n"
                    f"Hence defaulting to {self.model_name}",
                    UserWarning,
                )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
        else:
            model_name = self.model_name

        # Get the embeddings
        user_message_vector = self.model(
            text=user_message, model_name=model_name, **kwds
        )

        return user_message_vector


class AsyncEmbeddingModel(object):
    """Asynchronus Base class for embedding generation
    Default embedding model that will be used is `text-embedding-ada-002`"""

    default_model = DEFUALT_EMBEDDING_MODEL
    model_class = MODEL_CLASS

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Embedding Models class with the required parameters.
        Args:
            - model_name (str, optional): Name of the model to be used. Default: "gpt-4o-mini"

            ### Authentication parameters by provider:

            #### OpenAI models
            - api_key (str): OpenAI API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

            #### AWS Bedrock models
            - aws_access_key (str): AWS access key.
            - aws_secret_key (str): AWS secret key.
            - aws_region (str): AWS region name.
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

            #### Sentence Transformers models
            - model_download_path (str, optional): Path to download the model. Default: "/home/ubuntu/projects/models/embedding_models"
            - device (str, optional): Device to run the model. Default: "cpu"
            - trust_remote_code (bool, optional): Trust remote code. Default: False
            - token (str, optional): Hugging Face API token

            #### Azure OpenAI models
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - use_azure_openai (bool, optional): Use Azure OpenAI model. Default: False

        Raises:
            ValueError: If the model is not supported
            KeyError: If required parameters are missing
            TypeError: If the type of the parameter is incorrect
            ImportError: If the required library is not installed
        Warns:
            UserWarning: If no model_name is provided, defaulting to `text-embedding-ada-002`
        """

        # Check if the model name is provided
        if not kwds.get("model_name"):
            warnings.warn(
                f"\nNo 'model_name' specified, hence defaulting to {self.default_model}",
                UserWarning,
            )

        self.model_name = kwds.get("model_name", self.default_model)

        # Check if the model is supported
        if kwds.get("use_azure_openai"):
            self.model_type = "AzureOpenAI"
        elif hf_repo_exists(
            repo_id=self.model_name,
            repo_type=kwds.get("repo_type"),
            token=kwds.get("token"),
        ):
            self.model_type = "SentenceTransformer"
        else:
            self.model_type = self.model_class.get(self.model_name)

        if not self.model_type:
            raise ValueError(
                f"Unsupported model: {self.model_name}\n"
                f"Supported models are:\n"
                f"- "
                + "\n- ".join(list(self.model_class.keys()))
                + "\n- All sentence-transformers models"
            )

        # Initialize the model
        model_handler = {
            "OpenAI": openai_embeddings.AsyncEmbed,
            "AWSBedrock": awsbedrock_embeddings.AsyncEmbed,
            "SentenceTransformer": stransformers_embeddings.AsyncEmbed,
            "AzureOpenAI": azureopenai_embeddings.AsyncEmbed,
        }
        self.model = model_handler.get(self.model_type)(**kwds)

    async def __call__(
        self, user_message: Union[str, List[str]], **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """Get embeddings for the given text(s).
        Args:
            user_message (Union[str, List[str]]): Input text or list of texts
            **kwargs: Additional keyword arguments for the embedding API
        Returns:
            (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        Raises:
            KeyError: If required parameters are missing
            TypeError: If the type of the parameter is incorrect
        Warns:
            UserWarning: If the model is not supported or if the model is not found in the model type
        """

        # Check if the model name is provided
        if kwds.get("model_name"):
            # Check if the model is supported in the model type class
            if self.model_class.get(kwds.get("model_name")) == self.model_type:
                model_name = kwds.get("model_name")
            elif self.model_type == "SentenceTransformer":
                warnings.warn(
                    f"\nFor using different sentence-transformers model: {kwds.get('model_name')}\n"
                    f"again reinitialize the EmbeddingModels class as currently {self.model_name} is already loaded"
                    f"Hence defaulting the model to {self.model_name}",
                    UserWarning,
                )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
            else:
                warnings.warn(
                    f"Unsupported model: {kwds.get('model_name')}\nSupported models are:"
                    f"\n- "
                    + "\n- ".join(list(self.model_class.keys()))
                    + "\n- All sentence-transformers models\n"
                    f"Hence defaulting to {self.model_name}",
                    UserWarning,
                )
                # Defaulting to the model_name that is already loaded
                model_name = self.model_name
        else:
            model_name = self.model_name

        # Get the embeddings
        user_message_vector = await self.model(
            text=user_message, model_name=model_name, **kwds
        )

        return user_message_vector



from typing import Any, List, Dict, Union
import json

from concurrent.futures import ThreadPoolExecutor
import asyncio

from orichain import error_explainer


class Embed(object):
    """
    Synchronus AWS Bedrock Embeddings class to get embeddings for the given text(s).
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize AWS Bedrock client and set up API key.

        Args:
            aws_access_key (str): access key
            aws_secret_key (str): api key
            aws_region (str): region name
            config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    total_max_attempts: Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
        """

        from botocore.config import Config

        # Check for required parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("config") and not isinstance(kwds.get("config"), Config):
            raise TypeError(
                "Invalid 'config' type detected:",
                type(kwds.get("config")),
                ", Please enter valid config using:\n'from botocore.config import Config'",
            )
        else:
            pass

        import boto3

        # Initialize AWS Bedrock client
        self.client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=kwds.get("aws_access_key"),
            aws_secret_access_key=kwds.get("aws_secret_key"),
            config=kwds.get("config")
            or Config(
                region_name=kwds.get("aws_region"),
                read_timeout=10,
                connect_timeout=2,
                retries={"total_max_attempts": 2},
                max_pool_connections=100,
            ),
        )

        # Set up accept and content type for different models that wil be used by respective models
        self.accept = {
            "amazon.titan-embed-text-v1": "application/json",
            "amazon.titan-embed-text-v2:0": "application/json",
            "cohere.embed-english-v3": "*/*",
            "cohere.embed-multilingual-v3": "*/*",
        }
        self.content_type = "application/json"

    def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            text (Union[str, List[str]]): Input text or list of texts
            model_name (str): Name of the embedding model to use
            input_type (str, optional): Type of input text. Default: "search_query"
            truncate (str, optional): Truncate the input text. Default: "NONE"
            dimensions (int, optional): Output dimensions. Default: 1024 (Output dimensions can be: 256, 512 and 1024)
            normalize (bool, optional): Normalize the output. Default: True (As recommended in docs for RAG)
            **kwargs: Additional keyword arguments for the embedding API

        Returns:
           (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            self.embedding_types = kwds.get("embedding_types")

            if isinstance(text, str):
                text = [text]

            # This can be removed, as all will be independent calls
            if "cohere" in model_name:
                if len(text) > 96:
                    return {
                        "error": 500,
                        "reason": "too many embedding requests being processed, lower the number",
                    }

            # creatings tasks
            tasks = []
            for sentence in text:
                if "cohere" in model_name:
                    if len(sentence) > 2048 and kwds.get("truncate", "NONE") == "NONE":
                        return {
                            "error": 500,
                            "reason": f"length of query is {len(sentence)}, please lower it to 2048",
                        }

                    body = {
                        "texts": [sentence],
                        "input_type": kwds.get("input_type", "search_query"),
                        "truncate": kwds.get("truncate", "NONE"),
                    }
                    if kwds.get("embedding_types"):
                        body.update({"embedding_types": [kwds.get("embedding_types")]})

                elif "amazon" in model_name:
                    if "v2" in model_name:
                        body = {
                            "inputText": sentence,
                            "dimensions": kwds.get(
                                "dimensions", 1024
                            ),  # Output dimensions can be: 256, 512 and 1024
                            "normalize": kwds.get("normalize", True),
                        }  # As recommended in docs for RAG
                    elif "v1" in model_name:
                        body = {"inputText": sentence}

                tasks.append((body, model_name))

            # Use ThreadPoolExecutor for concurrent execution
            with ThreadPoolExecutor() as executor:
                embeddings = list(executor.map(self.worker, tasks))

            # check for errors
            for embedding in embeddings:
                if isinstance(embedding, Dict):
                    return embedding

            # return flattened list if only one embedding is present
            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def worker(self, task):
        """
        Worker function to execute in threads.
        Args:
            task (Tuple): Task to execute
        Returns:
            (Union[List[Union[float, int]], Dict]): Embeddings or error information
        """
        body, model_id = task
        return self._generate_embeddings(body=body, model_id=model_id)

    def _generate_embeddings(
        self, body: Dict, model_id: str
    ) -> Union[List[Union[float, int]], Dict]:
        """
        Generate embeddings for the given text.
        Args:
            body (Dict): Request body
            model_id (str): Model ID
        Returns:
            (Union[List[Union[float, int]], Dict]): Embeddings or error information
        """
        try:
            # Invoke model in a separate thread
            response = self.client.invoke_model(
                body=json.dumps(body),
                modelId=model_id,
                accept=self.accept.get(model_id),
                contentType=self.content_type,
            )

            # Extract embeddings from response
            response_body = json.loads(response.get("body").read())
            if "embeddings" in response_body:
                if self.embedding_types in response_body["embeddings"]:
                    return response_body["embeddings"].get(self.embedding_types)[0]
                else:
                    return response_body["embeddings"][0]
            else:
                return response_body["embedding"]

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncEmbed(object):
    """
    Asynchronus AWS Bedrock Embeddings class to get embeddings for the given text(s).
    """

    def __init__(self, **kwds: Any) -> None:
        """
        Initialize AWS Bedrock client and set up API key.

        Args:
            aws_access_key (str): access key
            aws_secret_key (str): api key
            aws_region (str): region name
            config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    total_max_attempts: Number of retries for the request. Default: 2

        Raises:
            KeyError: If required parameters are not provided.
        """

        from botocore.config import Config

        # Check for required parameters
        if not kwds.get("aws_access_key"):
            raise KeyError("Required 'aws_access_key' not found")
        elif not kwds.get("aws_secret_key"):
            raise KeyError("Required 'aws_secret_key' not found")
        elif not kwds.get("aws_region"):
            raise KeyError("Required aws_region not found")
        elif kwds.get("config") and not isinstance(kwds.get("config"), Config):
            raise TypeError(
                "Invalid 'config' type detected:",
                type(kwds.get("config")),
                ", Please enter valid config using:\n'from botocore.config import Config'",
            )
        else:
            pass

        import boto3

        # Initialize AWS Bedrock client
        self.client = boto3.client(
            service_name="bedrock-runtime",
            aws_access_key_id=kwds.get("aws_access_key"),
            aws_secret_access_key=kwds.get("aws_secret_key"),
            config=kwds.get("config")
            or Config(
                region_name=kwds.get("aws_region"),
                read_timeout=10,
                connect_timeout=2,
                retries={"total_max_attempts": 2},
                max_pool_connections=100,
            ),
        )

        # Set up accept and content type for different models that wil be used by respective models
        self.accept = {
            "amazon.titan-embed-text-v1": "application/json",
            "amazon.titan-embed-text-v2:0": "application/json",
            "cohere.embed-english-v3": "*/*",
            "cohere.embed-multilingual-v3": "*/*",
        }
        self.content_type = "application/json"

    async def __call__(
        self, text: Union[str, List[str]], model_name: str, **kwds: Any
    ) -> Union[List[float], List[List[float]], Dict]:
        """
        Get embeddings for the given text(s).

        Args:
            text (Union[str, List[str]]): Input text or list of texts
            model_name (str): Name of the embedding model to use
            **kwargs: Additional keyword arguments for the embedding API

        Returns:
           (Union[List[float], List[List[float]], Dict[str, Any]]): Embeddings or error information
        """
        try:
            self.embedding_types = kwds.get("embedding_types")

            if isinstance(text, str):
                text = [text]

            # This can be removed, as all will be independent calls
            if "cohere" in model_name:
                if len(text) > 96:
                    return {
                        "error": 500,
                        "reason": "too many embedding requests being processed, lower the number",
                    }

            # creatings tasks
            tasks = []
            for sentence in text:
                if "cohere" in model_name:
                    if len(sentence) > 2048 and kwds.get("truncate", "NONE") == "NONE":
                        return {
                            "error": 500,
                            "reason": f"length of query is {len(sentence)}, please lower it to 2048",
                        }

                    body = {
                        "texts": [sentence],
                        "input_type": kwds.get("input_type", "search_query"),
                        "truncate": kwds.get("truncate", "NONE"),
                    }
                    if kwds.get("embedding_types"):
                        body.update({"embedding_types": [kwds.get("embedding_types")]})

                elif "amazon" in model_name:
                    if "v2" in model_name:
                        body = {
                            "inputText": sentence,
                            "dimensions": kwds.get(
                                "dimensions", 1024
                            ),  # Output dimensions can be: 256, 512 and 1024
                            "normalize": kwds.get("normalize", True),
                        }  # As recommended in docs for RAG
                    elif "v1" in model_name:
                        body = {"inputText": sentence}

                tasks.append(self._generate_embeddings(body=body, model_id=model_name))

            # running tasks
            embeddings = await asyncio.gather(*tasks)

            # check for errors
            for embedding in embeddings:
                if isinstance(embedding, Dict):
                    return embedding

            # return flattened list if only one embedding is present
            if len(embeddings) == 1:
                embeddings = embeddings[0]

            return embeddings

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def _generate_embeddings(
        self, body: Dict, model_id: str
    ) -> Union[List[Union[float, int]], Dict]:
        """
        Generate embeddings for the given text.
        Args:
            body (Dict): Request body
            model_id (str): Model ID
        Returns:
            (Union[List[Union[float, int]], Dict]): Embeddings or error information
        """
        try:
            # Invoke model in a separate thread
            response = await asyncio.to_thread(
                self.client.invoke_model,
                body=json.dumps(body),
                modelId=model_id,
                accept=self.accept.get(model_id),
                contentType=self.content_type,
            )

            # Extract embeddings from response
            response_body = json.loads(response.get("body").read())
            if "embeddings" in response_body:
                if self.embedding_types in response_body["embeddings"]:
                    return response_body["embeddings"].get(self.embedding_types)[0]
                else:
                    return response_body["embeddings"][0]
            else:
                return response_body["embedding"]

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Any, List, Union, Dict

from orichain import error_explainer

import asyncio


class DataBase(object):
    """
    Synchronous way to interact with the Chromadb knowledge base
    """

    def __init__(self, **kwds) -> None:
        """Initializes the Chromadb client.
        Args:
            path (str, optional): Path to the Chromadb Defualt: `/home/ubuntu/projects/chromadb`
            collection_name (str): Chromadb collection name
        Raises:
            KeyError: If required parameters are not found
        """
        if not kwds.get("collection_name"):
            raise KeyError("Required `collection_name` not found")

        ## WILL NOT BE USED as we are not passing text directly while retrieving the chunks in query and get methods, keep it here just for reference
        # elif not kwds.get("embedding_function"):
        #     warnings.warn(
        #         "‘embedding_function’ has not been defined while initializing the KnowledgeBase class.\n"
        #         "You are using ChromaDB; if a collection has been created with an embedding model different from the "
        #         "default embedding model (`all-MiniLM-L6-v2`) that ChromaDB uses, YOU NEED TO PASS 'embedding_function' LIKE THIS:\n"
        #         "------------------------------------------------------\n"
        #         "class MyEmbeddingFunction(chromadb.EmbeddingFunction):\n"
        #         "\tdef __call__(self, input: chromadb.Documents) -> chromadb.Embeddings:\n"
        #         "\t\tembeddings = []\n"
        #         "\t\tfor text in input:\n"
        #         "\t\t\tembeddings.append(embedding_model(text=text))\n"
        #         "\t\treturn embeddings",
        #         UserWarning,
        #     )

        else:
            pass

        try:
            # Importing pysqlite3 as sqlite3
            __import__("pysqlite3")
            import sys

            sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")

            import chromadb

            # TODO Need to create a separate class for HTTPClient but now we are using PersistentClient

            # Initializing the Chromadb client
            self.client = chromadb.PersistentClient(
                path=kwds.get("path", "/home/ubuntu/projects/chromadb")
            )

            # Getting the collection
            self.collection = self.client.get_collection(
                name=kwds.get("collection_name")
            )
        except Exception as e:
            error_explainer(e)

    def __call__(
        self,
        user_message_vector: List[Union[int, float]],
        num_of_chunks: int,
        **kwds: Any,
    ) -> Any:
        """Retrieves the chunks from the knowledge base using query method
        Args:
            user_message_vector (List[Union[int, float]]): Embedding of text
            num_of_chunks (int): Number of chunks to retrieve
        Returns:
            Dict: Result of retrieving the chunks
        """
        try:
            # If collection_name is provided, get the collection
            if kwds.get("collection_name"):
                collection = self.client.get_collection(
                    name=kwds.get("collection_name")
                )
            else:
                collection = self.collection

            # Querying the collection
            chunks = collection.query(
                query_embeddings=user_message_vector,
                n_results=num_of_chunks,
                where=kwds.get("where"),
                where_document=kwds.get("where_document"),
                include=kwds.get("include")
                if kwds.get("include")
                else ["metadatas", "documents"],
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def fetch(self, ids: List[str], **kwds: Any) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base using get method
        Args:
            ids (List[str]): List of ids to fetch
        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # If collection_name is provided, get the collection
            if kwds.get("collection_name"):
                collection = self.client.get_collection(
                    name=kwds.get("collection_name")
                )
            else:
                collection = self.collection

            # Fetching the chunks based on the ids
            chunks = collection.get(
                ids=ids,
                limit=kwds.get("limit"),
                offset=kwds.get("offset"),
                where=kwds.get("where"),
                where_document=kwds.get("where_document"),
                include=kwds.get("include")
                if kwds.get("include")
                else ["metadatas", "documents"],
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncDataBase(object):
    """
    Asynchronous way to interact with the Chromadb knowledge base
    """

    def __init__(self, **kwds) -> None:
        """Initializes the Chromadb client.
        Args:
            path (str): Path to the Chromadb
            collection_name (str): Chromadb collection name
        Raises:
            KeyError: If required parameters are not found
        """
        if not kwds.get("collection_name"):
            raise KeyError("Required `collection_name` not found")

        ## WILL NOT BE USED as we are not passing text directly while retrieving the chunks in query and get methods, keep it here just for reference
        # elif not kwds.get("embedding_function"):
        #     warnings.warn(
        #         "‘embedding_function’ has not been defined while initializing the KnowledgeBase class.\n"
        #         "You are using ChromaDB; if a collection has been created with an embedding model different from the "
        #         "default embedding model (`all-MiniLM-L6-v2`) that ChromaDB uses, YOU NEED TO PASS 'embedding_function' LIKE THIS:\n"
        #         "------------------------------------------------------\n"
        #         "class MyEmbeddingFunction(chromadb.EmbeddingFunction):\n"
        #         "\tdef __call__(self, input: chromadb.Documents) -> chromadb.Embeddings:\n"
        #         "\t\tembeddings = []\n"
        #         "\t\tfor text in input:\n"
        #         "\t\t\tembeddings.append(embedding_model(text=text))\n"
        #         "\t\treturn embeddings",
        #         UserWarning,
        #     )

        else:
            pass

        try:
            # Importing pysqlite3 as sqlite3
            __import__("pysqlite3")
            import sys

            sys.modules["sqlite3"] = sys.modules.pop("pysqlite3")

            import chromadb

            # TODO Need to create a separate class for HTTPClient but now we are using PersistentClient

            # Initializing the Chromadb client
            self.client = chromadb.PersistentClient(
                path=kwds.get("path", "/home/ubuntu/projects/chromadb")
            )

            # Getting the collection
            self.collection = self.client.get_collection(
                name=kwds.get("collection_name")
            )
        except Exception as e:
            error_explainer(e)

    async def __call__(
        self,
        user_message_vector: List[Union[int, float]],
        num_of_chunks: int,
        **kwds: Any,
    ) -> Any:
        """Retrieves the chunks from the knowledge base using query method
        Args:
            user_message_vector (List[Union[int, float]]): Embedding of text
            num_of_chunks (int): Number of chunks to retrieve
        Returns:
            Dict: Result of retrieving the chunks
        """
        try:
            # If collection_name is provided, get the collection
            if kwds.get("collection_name"):
                collection = self.client.get_collection(
                    name=kwds.get("collection_name")
                )
            else:
                collection = self.collection

            # Querying the collection
            chunks = await asyncio.to_thread(
                collection.query,
                query_embeddings=user_message_vector,
                n_results=num_of_chunks,
                where=kwds.get("where"),
                where_document=kwds.get("where_document"),
                include=kwds.get("include")
                if kwds.get("include")
                else ["metadatas", "documents"],
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def fetch(self, ids: List[str], **kwds: Any) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base using get method
        Args:
            ids (List[str]): List of ids to fetch
        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # If collection_name is provided, get the collection
            if kwds.get("collection_name"):
                collection = self.client.get_collection(
                    name=kwds.get("collection_name")
                )
            else:
                collection = self.collection

            # Fetching the chunks based on the ids
            chunks = await asyncio.to_thread(
                collection.get,
                ids=ids,
                limit=kwds.get("limit"),
                offset=kwds.get("offset"),
                where=kwds.get("where"),
                where_document=kwds.get("where_document"),
                include=kwds.get("include")
                if kwds.get("include")
                else ["metadatas", "documents"],
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Any, List, Union, Optional, Dict
import warnings

from orichain.knowledge_base import pinecone_knowledgbase, chromadb_knowledgebase
from orichain import error_explainer

DEFAULT_KNOWLEDGE_BASE = "pinecone"


class KnowledgeBase(object):
    """
    Synchronous way to retrieve chunks from the knowledge base
    """

    default_knowledge_base = DEFAULT_KNOWLEDGE_BASE

    def __init__(self, vector_db_type: Optional[str], **kwds: Any) -> None:
        """Initializes the knowledge base.

        Args:
            vector_db_type (str, optional): Type of knowledge base. Default: pinecone

        Authentication parameters by provider:

            Pinecone:
            api_key (str): Pinecone API key
            index_name (str): Pinecone index name
            namespace (str): Pinecone namespace

            ChromaDB:
            collection_name (str): ChromaDB collection name
            path (str, optional): Path to the ChromaDB database Default: `/home/ubuntu/projects/chromadb`

        Raises:
            ValueError: If the knowledge base type is not supported
            KeyError: If the required params is not found

        Warns:
            UserWarning: If the knowledge base type is not defined Default: pinecone
        """
        try:
            # Dictionary mapping vector database types to their respective handler classes
            knowledge_base_handler = {
                "pinecone": pinecone_knowledgbase.DataBase,
                "chromadb": chromadb_knowledgebase.DataBase,
            }

            # If no vector_db_type is provided, default to pinecone
            if not vector_db_type:
                warnings.warn(
                    f"\nKnowledge base type not defined hence defaulting to \
                    {self.default_knowledge_base}",
                    UserWarning,
                )
                self.vector_db_type = self.default_knowledge_base
            # If vector_db_type is not supported, raise a ValueError
            elif vector_db_type not in list(knowledge_base_handler.keys()):
                raise ValueError(
                    f"\nUnsupported knowledge base: {self.model_name}\nSupported knowledge bases are:"
                    f"\n- " + "\n- ".join(list(knowledge_base_handler.keys()))
                )
            else:
                self.vector_db_type = vector_db_type

            # Initialize the knowledge base handler
            self.retriver = knowledge_base_handler.get(
                vector_db_type, self.default_knowledge_base
            )(**kwds)

        except Exception as e:
            error_explainer(e)

    def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves the chunks from the knowledge base
        Args:
            num_of_chunks (int): Number of chunks to retrieve
            user_message_vector (Optional[List[Union[int, float]]]): Embedding of text. Defaults to None.

        Returns:
            Dict: Result of retrieving the chunks

        Raises:
            ValueError: If `user_message_vector` is needed except for pinecone but if ids are also not provided for pinecone this error will be raised
            KeyError: If required `namespace` is not found for pinecone
        """
        try:
            if not user_message_vector and not self.vector_db_type == "pinecone":
                raise ValueError("`user_message_vector` is needed except for pinecone")

            chunks = self.retriver(
                user_message_vector=user_message_vector,
                num_of_chunks=num_of_chunks,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def fetch(
        self,
        ids: List[str],
        **kwds: Any,
    ) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base
        Args:
            ids (List[str]): List of ids to fetch

        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = self.retriver.fetch(
                ids=ids,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}


class AsyncKnowledgeBase(object):
    """
    Asynchronous way to retrieve chunks from the knowledge base
    """

    default_knowledge_base = DEFAULT_KNOWLEDGE_BASE

    def __init__(self, vector_db_type: Optional[str], **kwds: Any) -> None:
        """Initializes the knowledge base.

        Args:
            vector_db_type (str, optional): Type of knowledge base. Default: pinecone

        Authentication parameters by provider:

            Pinecone:
            api_key (str): Pinecone API key
            index_name (str): Pinecone index name
            namespace (str): Pinecone namespace

            ChromaDB:
            collection_name (str): ChromaDB collection name
            path (str, optional): Path to the ChromaDB database Default: `/home/ubuntu/projects/chromadb`

        Raises:
            ValueError: If the knowledge base type is not supported
            KeyError: If the required params is not found

        Warns:
            UserWarning: If the knowledge base type is not defined Default: pinecone
        """
        try:
            # Dictionary mapping vector database types to their respective handler classes
            knowledge_base_handler = {
                "pinecone": pinecone_knowledgbase.AsyncDataBase,
                "chromadb": chromadb_knowledgebase.AsyncDataBase,
            }

            # If no vector_db_type is provided, default to pinecone
            if not vector_db_type:
                warnings.warn(
                    f"\nKnowledge base type not defined hence defaulting to \
                    {self.default_knowledge_base}",
                    UserWarning,
                )
                self.vector_db_type = self.default_knowledge_base
            # If vector_db_type is not supported, raise a ValueError
            elif vector_db_type not in list(knowledge_base_handler.keys()):
                raise ValueError(
                    f"\nUnsupported knowledge base: {self.model_name}\nSupported knowledge bases are:"
                    f"\n- " + "\n- ".join(list(knowledge_base_handler.keys()))
                )
            else:
                self.vector_db_type = vector_db_type

            # Initialize the knowledge base handler
            self.retriver = knowledge_base_handler.get(
                vector_db_type, self.default_knowledge_base
            )(**kwds)

        except Exception as e:
            error_explainer(e)

    async def __call__(
        self,
        num_of_chunks: int,
        user_message_vector: Optional[List[Union[int, float]]] = None,
        **kwds: Any,
    ) -> Dict:
        """Retrieves the chunks from the knowledge base
        Args:
            num_of_chunks (int): Number of chunks to retrieve
            user_message_vector (Optional[List[Union[int, float]]]): Embedding of text. Defaults to None.

        Returns:
            Dict: Result of retrieving the chunks

        Raises:
            ValueError: If `user_message_vector` is needed except for pinecone but if ids are also not provided for pinecone this error will be raised
            KeyError: If required `namespace` is not found for pinecone
        """
        try:
            if not user_message_vector and not self.vector_db_type == "pinecone":
                raise ValueError("`user_message_vector` is needed except for pinecone")

            chunks = await self.retriver(
                user_message_vector=user_message_vector,
                num_of_chunks=num_of_chunks,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def fetch(
        self,
        ids: List[str],
        **kwds: Any,
    ) -> Dict:
        """Fetches the chunks based on the ids from the knowledge base
        Args:
            ids (List[str]): List of ids to fetch

        Returns:
            Dict: Result of fetching the chunks
        """
        try:
            # Fetching the chunks based on the ids
            chunks = await self.retriver.fetch(
                ids=ids,
                **kwds,
            )

            return chunks

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}



from typing import Any, Optional, List, Dict, Generator, AsyncGenerator
import warnings
import json
from fastapi import Request

from orichain import error_explainer

from orichain.llm import (
    openai_llm,
    anthropicbedrock_llm,
    anthropic_llm,
    awsbedrock_llm,
    azureopenai_llm,
)

DEFAULT_MODEL = "gpt-4o-mini"
MODEL_CLASS = {
    "gpt-4o": "OpenAI",
    "gpt-4-turbo": "OpenAI",
    "gpt-4-turbo-preview": "OpenAI",
    "gpt-4o-mini": "OpenAI",
    "gpt-4": "OpenAI",
    "gpt-4.1": "OpenAI",
    "gpt-4.1-mini": "OpenAI",
    "gpt-4.1-nano": "OpenAI",
    "anthropic.claude-3-haiku-20240307-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-haiku-20240307-v1:0": "AnthropicAWSBedrock",
    "us-gov.anthropic.claude-3-haiku-20240307-v1:0": "AnthropicAWSBedrock",
    "eu.anthropic.claude-3-haiku-20240307-v1:0": "AnthropicAWSBedrock",
    "apac.anthropic.claude-3-haiku-20240307-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-5-haiku-20241022-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-5-haiku-20241022-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-sonnet-20240229-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-sonnet-20240229-v1:0": "AnthropicAWSBedrock",
    "eu.anthropic.claude-3-sonnet-20240229-v1:0": "AnthropicAWSBedrock",
    "apac.anthropic.claude-3-sonnet-20240229-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-5-sonnet-20240620-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-5-sonnet-20240620-v1:0": "AnthropicAWSBedrock",
    "us-gov.anthropic.claude-3-5-sonnet-20240620-v1:0": "AnthropicAWSBedrock",
    "eu.anthropic.claude-3-5-sonnet-20240620-v1:0": "AnthropicAWSBedrock",
    "apac.anthropic.claude-3-5-sonnet-20240620-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-5-sonnet-20241022-v2:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-5-sonnet-20241022-v2:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-7-sonnet-20250219-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-7-sonnet-20250219-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-sonnet-4-20250514-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-sonnet-4-20250514-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-3-opus-20240229-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-3-opus-20240229-v1:0": "AnthropicAWSBedrock",
    "anthropic.claude-opus-4-20250514-v1:0": "AnthropicAWSBedrock",
    "us.anthropic.claude-opus-4-20250514-v1:0": "AnthropicAWSBedrock",
    "claude-3-haiku-20240307": "Anthropic",
    "claude-3-5-haiku-latest": "Anthropic",
    "claude-3-sonnet-20240229": "Anthropic",
    "claude-3-5-sonnet-latest": "Anthropic",
    "claude-3-7-sonnet-latest": "Anthropic",
    "claude-sonnet-4-20250514": "Anthropic",
    "claude-3-opus-latest": "Anthropic",
    "claude-opus-4-20250514": "Anthropic",
    "cohere.command-text-v14": "AWSBedrock",
    "cohere.command-light-text-v14": "AWSBedrock",
    "cohere.command-r-v1:0": "AWSBedrock",
    "cohere.command-r-plus-v1:0": "AWSBedrock",
    "meta.llama3-8b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-70b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-1-8b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-1-8b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-1-70b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-1-70b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-1-405b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-2-1b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-2-1b-instruct-v1:0": "AWSBedrock",
    "eu.meta.llama3-2-1b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-2-3b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-2-3b-instruct-v1:0": "AWSBedrock",
    "eu.meta.llama3-2-3b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-2-11b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-2-11b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-2-90b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-2-90b-instruct-v1:0": "AWSBedrock",
    "meta.llama3-3-70b-instruct-v1:0": "AWSBedrock",
    "us.meta.llama3-3-70b-instruct-v1:0": "AWSBedrock",
    "mistral.mistral-7b-instruct-v0:2": "AWSBedrock",
    "mistral.mixtral-8x7b-instruct-v0:1": "AWSBedrock",
    "mistral.mistral-large-2402-v1:0": "AWSBedrock",
    "mistral.mistral-large-2407-v1:0": "AWSBedrock",
    "mistral.mistral-small-2402-v1:0": "AWSBedrock",
    "amazon.titan-text-express-v1": "AWSBedrock",
    "amazon.titan-text-lite-v1": "AWSBedrock",
    "amazon.titan-text-premier-v1:0": "AWSBedrock",
    "amazon.nova-pro-v1:0": "AWSBedrock",
    "us.amazon.nova-pro-v1:0": "AWSBedrock",
    "amazon.nova-lite-v1:0": "AWSBedrock",
    "us.amazon.nova-lite-v1:0": "AWSBedrock",
    "amazon.nova-micro-v1:0": "AWSBedrock",
    "us.amazon.nova-micro-v1:0": "AWSBedrock",
}


class LLM(object):
    """Synchronous Language Model class for interacting with various LLM providers.

    This class provides a unified interface to interact with different language models
    from providers such as OpenAI, AWS Bedrock, Anthropic, and Azure OpenAI.
    """

    default_model = DEFAULT_MODEL
    model_class = MODEL_CLASS

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Language Model class with the required parameters.

        Args:
            - model_name (str, optional): Name of the model to be used. Default: "gpt-4o-mini"

            ### Authentication parameters by provider:

            #### OpenAI models
            - api_key (str): OpenAI API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

            #### AWS Bedrock models
            - aws_access_key (str): AWS access key.
            - aws_secret_key (str): AWS secret key.
            - aws_region (str): AWS region name.
            - prompt_caching (bool, optional): Whether to use prompt caching. Default: True
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

            #### Anthropic models
            - api_key (str): Anthropic API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - prompt_caching (bool, optional): Whether to use prompt caching. Default: True

            #### Azure OpenAI models
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - use_azure_openai (bool, optional): Whether to use Azure OpenAI service. Default: False

        Raises:
            ValueError: If an unsupported model is specified.
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter.

        Warns:
            UserWarning: If the model name is not provided, it defaults to the default model.
        """

        # Set model name, defaulting if not provided
        if not kwds.get("model_name"):
            if kwds.get("use_azure_openai"):
                warnings.warn(
                    f"\nNo 'model_name' specified, hence defaulting to {self.default_model} (Azure OpenAI)",
                    UserWarning,
                )
            else:
                warnings.warn(
                    f"\nNo 'model_name' specified, hence defaulting to {self.default_model} (OpenAI)",
                    UserWarning,
                )
        self.model_name = kwds.get("model_name", self.default_model)

        # Check if Azure OpenAI model is to be used
        if kwds.get("use_azure_openai"):
            self.model_type = "AzureOpenAI"
        else:
            # Determine model type
            self.model_type = self.model_class.get(self.model_name)

        if not self.model_type:
            raise ValueError(
                f"\nUnsupported model: {self.model_name}\nSupported models are:"
                f"\n- " + "\n- ".join(list(self.model_class.keys()))
            )

        # Initialize model handler and map model types to their respective handler classes
        model_handler = {
            "OpenAI": openai_llm.Generate,
            "AWSBedrock": awsbedrock_llm.Generate,
            "AnthropicAWSBedrock": anthropicbedrock_llm.Generate,
            "Anthropic": anthropic_llm.Generate,
            "AzureOpenAI": azureopenai_llm.Generate,
        }

        # Initialize the appropriate model handler
        self.model = model_handler.get(self.model_type)(**kwds)

    def __call__(
        self,
        user_message: str,
        matched_sentence: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        extra_metadata: Optional[Dict] = None,
        do_json: bool = False,
        **kwds: Any,
    ) -> Dict:
        """Generate a synchronous response from the language model.

        Args:
            user_message (str): The user's input message.
            matched_sentence (List[str], optional): List of matched sentences for context.
            system_prompt (str, optional): System prompt to guide the model's behavior.
            chat_hist (List[Dict[str, str]], optional): Chat history for context.
            sampling_paras (Dict, optional): Parameters for sampling (temperature, top_p, etc.).
            extra_metadata (Dict, optional): Additional metadata to include in the response.
            do_json (bool, optional): Whether to return a JSON response. Default: False.
            **kwds: Additional keyword arguments to pass to the model.

        Returns:
            Dict: The model's response with metadata.
        """
        try:
            # Handle model switching if a different model is specified in kwds
            if self._model_n_model_type_validator(**kwds):
                model_name = kwds.pop("model_name", self.model_name)
            else:
                model_name = self.model_name

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}
            extra_metadata = extra_metadata or {}

            # Generate the response
            result = self.model(
                model_name=model_name,
                user_message=user_message,
                system_prompt=system_prompt,
                chat_hist=chat_hist,
                sampling_paras=sampling_paras,
                do_json=do_json,
                **kwds,
            )

            # Add user message and matched sentence to the response
            if "error" not in result:
                result.update({"message": user_message})
                if matched_sentence:
                    result.update({"matched_sentence": matched_sentence})
                # Add extra metadata to the response
                if extra_metadata:
                    result["metadata"].update(extra_metadata)

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    def stream(
        self,
        user_message: str,
        matched_sentence: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
        chat_hist: List = None,
        sampling_paras: Optional[Dict] = None,
        extra_metadata: Optional[Dict] = None,
        do_json: bool = False,
        do_sse: bool = True,
        **kwds: Any,
    ) -> Generator:
        """Stream responses from the language model.

        Args:
            user_message (str): The user's input message.
            matched_sentence (List[str], optional): List of matched sentences for context.
            system_prompt (str, optional): System prompt to guide the model's behavior.
            chat_hist (List[Dict[str, str]], optional): Chat history for context.
            sampling_paras (Dict, optional): Parameters for sampling (temperature, top_p, etc.).
            extra_metadata (Dict, optional): Additional metadata to include in the response.
            do_json (bool, optional): Whether to return JSON responses. Default: False.
            do_sse (bool, optional): Whether to format responses as Server-Sent Events. Default: True.
            **kwds: Additional keyword arguments to pass to the model.

        Yields:
            Generator: Stream of responses from the language model.
        """
        try:
            # Handle model switching if a different model is specified in kwds
            if self._model_n_model_type_validator(**kwds):
                model_name = kwds.get("model_name", self.model_name)
            else:
                model_name = self.model_name

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}
            extra_metadata = extra_metadata or {}

            # Stream responses from the model
            result = self.model.streaming(
                model_name=model_name,
                user_message=user_message,
                system_prompt=system_prompt,
                chat_hist=chat_hist,
                sampling_paras=sampling_paras,
                do_json=do_json,
                **kwds,
            )

            # Process each chunk in the stream
            for chunk in result:
                if isinstance(chunk, str):
                    if do_sse:
                        yield self._format_sse(chunk, event="text")
                    else:
                        yield chunk
                elif isinstance(chunk, Dict):
                    if "error" not in chunk:
                        chunk.update(
                            {
                                "message": user_message,
                            }
                        )
                        if matched_sentence:
                            chunk.update({"matched_sentence": matched_sentence})
                        if extra_metadata:
                            chunk["metadata"].update(extra_metadata)
                    if do_sse:
                        yield self._format_sse(chunk, event="body")
                    else:
                        yield chunk

        except Exception as e:
            error_explainer(e)
            yield self._format_sse({"error": 500, "reason": str(e)}, event="body")

    def _format_sse(self, data: Any, event=None) -> str:
        """Format data for Server-Sent Events (SSE).

        Args:
            data (Any): The data to format.
            event (str, optional): The event type.

        Returns:
            str: Formatted SSE message.
        """
        msg = f"data: {json.dumps(data)}\n\n"

        if event is not None:
            msg = f"event: {event}\n{msg}"

        return msg

    def _model_n_model_type_validator(self, **kwds: Any) -> bool:
        """Validate if the requested model is compatible with the current model type.

        Args:
            **kwds: Keyword arguments that may contain a 'model_name'.

        Returns:
            bool: True if the model is compatible, False otherwise.
        """

        if self.model_type == "AzureOpenAI":
            return True
        elif kwds.get("model_name"):
            if self.model_class.get(kwds.get("model_name")) == self.model_type:
                return True
            elif (
                self.model_class.get(kwds.get("model_name"))
                and self.model_class.get(kwds.get("model_name")) != self.model_type
            ):
                warnings.warn(
                    f"{kwds.get('model_name')} is a supported model but "
                    f"does not belong to {self.model_type}, again reinitialize the "
                    f"LLM class with {self.model_class.get(kwds.get('model_name'))} model class. "
                    f"Hence defaulting the model to {self.model_name}",
                    UserWarning,
                )
                return False
            else:
                warnings.warn(
                    f"Unsupported model: {kwds.get('model_name')}\nSupported models are:"
                    f"\n- "
                    + "\n- ".join(list(self.model_class.keys()))
                    + "\n- All sentence-transformers models\n"
                    f"Hence defaulting to {self.model_name}",
                    UserWarning,
                )
                return False
        else:
            return False


class AsyncLLM(object):
    """Asynchronous Language Model class for interacting with various LLM providers.

    This class provides a unified interface to interact with different language models
    from providers such as OpenAI, AWS Bedrock, Anthropic, and Azure OpenAI.
    """

    default_model = DEFAULT_MODEL
    model_class = MODEL_CLASS

    def __init__(self, **kwds: Any) -> None:
        """Initialize the Language Model class with the required parameters.

        Args:
            - model_name (str, optional): Name of the model to be used. Default: "gpt-4o-mini"

            ### Authentication parameters by provider:

            #### OpenAI models
            - api_key (str): OpenAI API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2

            #### AWS Bedrock models
            - aws_access_key (str): AWS access key.
            - aws_secret_key (str): AWS secret key.
            - aws_region (str): AWS region name.
            - prompt_caching (bool, optional): Whether to use prompt caching. Default: True
            - config (Config, optional):
                - connect_timeout (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to make a connection. Default: 60
                - read_timeout: (float or int, optional): The time in seconds till a timeout exception is
                thrown when attempting to read from a connection. Default: 60
                - region_name (str, optional): region name Note: If specifing config you need to still pass region_name even if you have already passed in aws_region
                - max_pool_connections: The maximum number of connections to keep in a connection pool. Defualt: 10
                - retries (Dict, optional):
                    - total_max_attempts: Number of retries for the request. Default: 2

            #### Anthropic models
            - api_key (str): Anthropic API key.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - prompt_caching (bool, optional): Whether to use prompt caching. Default: True

            #### Azure OpenAI models
            - api_key (str): Azure OpenAI API key.
            - azure_endpoint (str): Azure OpenAI endpoint.
            - api_version (str): Azure OpenAI API version.
            - timeout (Timeout, optional): Request timeout parameter like connect, read, write. Default: 60.0, 5.0, 10.0, 2.0
            - max_retries (int, optional): Number of retries for the request. Default: 2
            - use_azure_openai (bool, optional): Whether to use Azure OpenAI service. Default: False

        Raises:
            ValueError: If an unsupported model is specified.
            KeyError: If required parameters are not provided.
            TypeError: If an invalid type is provided for a parameter.

        Warns:
            UserWarning: If the model name is not provided, it defaults to the default model.
        """

        # Set model name, defaulting if not provided
        if not kwds.get("model_name"):
            warnings.warn(
                f"\nNo 'model_name' specified, hence defaulting to {self.default_model} (OpenAI)",
                UserWarning,
            )
        self.model_name = kwds.get("model_name", self.default_model)

        # Determine model type
        self.model_type = self.model_class.get(self.model_name)

        # Check if Azure OpenAI model is to be used
        if kwds.get("use_azure_openai"):
            self.model_type = "AzureOpenAI"

        if not self.model_type:
            raise ValueError(
                f"\nUnsupported model: {self.model_name}\nSupported models are:"
                f"\n- " + "\n- ".join(list(self.model_class.keys()))
            )

        # Initialize model handler and map model types to their respective handler classes
        model_handler = {
            "OpenAI": openai_llm.AsyncGenerate,
            "AWSBedrock": awsbedrock_llm.AsyncGenerate,
            "AnthropicAWSBedrock": anthropicbedrock_llm.AsyncGenerate,
            "Anthropic": anthropic_llm.AsyncGenerate,
            "AzureOpenAI": azureopenai_llm.AsyncGenerate,
        }

        # Initialize the appropriate model handler
        self.model = model_handler.get(self.model_type)(**kwds)

    async def __call__(
        self,
        user_message: str,
        request: Optional[Request] = None,
        matched_sentence: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
        chat_hist: Optional[List[Dict[str, str]]] = None,
        sampling_paras: Optional[Dict] = None,
        extra_metadata: Optional[Dict] = None,
        do_json: bool = False,
        **kwds: Any,
    ) -> Dict:
        """Generate a synchronous response from the language model.

        Args:
            user_message (str): The user's input message.
            request (Request, optional): FastAPI request object for cancellation detection.
            matched_sentence (List[str], optional): List of matched sentences for context.
            system_prompt (str, optional): System prompt to guide the model's behavior.
            chat_hist (List[Dict[str, str]], optional): Chat history for context.
            sampling_paras (Dict, optional): Parameters for sampling (temperature, top_p, etc.).
            extra_metadata (Dict, optional): Additional metadata to include in the response.
            do_json (bool, optional): Whether to return a JSON response. Default: False.
            **kwds: Additional keyword arguments to pass to the model.

        Returns:
            Dict: The model's response with metadata.
        """
        try:
            # Handle model switching if a different model is specified in kwds
            if await self._model_n_model_type_validator(**kwds):
                model_name = kwds.pop("model_name", self.model_name)
            else:
                model_name = self.model_name

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}
            extra_metadata = extra_metadata or {}

            # Check if request is disconnected
            if request and await request.is_disconnected():
                return {"error": 400, "reason": "request aborted by user"}

            # Generate the response
            result = await self.model(
                request=request,
                model_name=model_name,
                user_message=user_message,
                system_prompt=system_prompt,
                chat_hist=chat_hist,
                sampling_paras=sampling_paras,
                do_json=do_json,
                **kwds,
            )

            # Add user message and matched sentence to the response
            if "error" not in result:
                result.update({"message": user_message})
                if matched_sentence:
                    result.update({"matched_sentence": matched_sentence})
                # Add extra metadata to the response
                if extra_metadata:
                    result["metadata"].update(extra_metadata)

            return result

        except Exception as e:
            error_explainer(e)
            return {"error": 500, "reason": str(e)}

    async def stream(
        self,
        user_message: str,
        request: Optional[Request] = None,
        matched_sentence: Optional[List[str]] = None,
        system_prompt: Optional[str] = None,
        chat_hist: List = None,
        sampling_paras: Optional[Dict] = None,
        extra_metadata: Optional[Dict] = None,
        do_json: bool = False,
        do_sse: bool = True,
        **kwds: Any,
    ) -> AsyncGenerator:
        """Stream responses from the language model.

        Args:
            user_message (str): The user's input message.
            request (Request, optional): FastAPI request object for cancellation detection.
            matched_sentence (List[str], optional): List of matched sentences for context.
            system_prompt (str, optional): System prompt to guide the model's behavior.
            chat_hist (List[Dict[str, str]], optional): Chat history for context.
            sampling_paras (Dict, optional): Parameters for sampling (temperature, top_p, etc.).
            extra_metadata (Dict, optional): Additional metadata to include in the response.
            do_json (bool, optional): Whether to return JSON responses. Default: False.
            do_sse (bool, optional): Whether to format responses as Server-Sent Events. Default: True.
            **kwds: Additional keyword arguments to pass to the model.

        Yields:
            AsyncGenerator: Stream of responses from the language model.
        """
        try:
            # Handle model switching if a different model is specified in kwds
            if await self._model_n_model_type_validator(**kwds):
                model_name = kwds.get("model_name", self.model_name)
            else:
                model_name = self.model_name

            # Default empty dictionaries
            sampling_paras = sampling_paras or {}
            extra_metadata = extra_metadata or {}

            # Check if the request has been disconnected
            if request and await request.is_disconnected():
                yield await self._format_sse(
                    {"error": 400, "reason": "request aborted by user"}, event="body"
                )
            else:
                # Stream responses from the model
                result = self.model.streaming(
                    request=request,
                    model_name=model_name,
                    user_message=user_message,
                    system_prompt=system_prompt,
                    chat_hist=chat_hist,
                    sampling_paras=sampling_paras,
                    do_json=do_json,
                    **kwds,
                )

                # Process each chunk in the stream
                async for chunk in result:
                    if isinstance(chunk, str):
                        if do_sse:
                            yield await self._format_sse(chunk, event="text")
                        else:
                            yield chunk
                    elif isinstance(chunk, Dict):
                        if "error" not in chunk:
                            chunk.update(
                                {
                                    "message": user_message,
                                }
                            )
                            if matched_sentence:
                                chunk.update({"matched_sentence": matched_sentence})
                            if extra_metadata:
                                chunk["metadata"].update(extra_metadata)
                        if do_sse:
                            yield await self._format_sse(chunk, event="body")
                        else:
                            yield chunk

        except Exception as e:
            error_explainer(e)
            yield await self._format_sse({"error": 500, "reason": str(e)}, event="body")

    async def _format_sse(self, data: Any, event=None) -> str:
        """Format data for Server-Sent Events (SSE).

        Args:
            data (Any): The data to format.
            event (str, optional): The event type.

        Returns:
            str: Formatted SSE message.
        """
        msg = f"data: {json.dumps(data)}\n\n"

        if event is not None:
            msg = f"event: {event}\n{msg}"

        return msg

    async def _model_n_model_type_validator(self, **kwds: Any) -> bool:
        """Validate if the requested model is compatible with the current model type.

        Args:
            **kwds: Keyword arguments that may contain a 'model_name'.

        Returns:
            bool: True if the model is compatible, False otherwise.
        """

        if self.model_type == "AzureOpenAI":
            return True
        elif kwds.get("model_name"):
            if self.model_class.get(kwds.get("model_name")) == self.model_type:
                return True
            elif (
                self.model_class.get(kwds.get("model_name"))
                and self.model_class.get(kwds.get("model_name")) != self.model_type
            ):
                warnings.warn(
                    f"{kwds.get('model_name')} is a supported model but "
                    f"does not belong to {self.model_type}, again reinitialize the "
                    f"LLM class with {self.model_class.get(kwds.get('model_name'))} model class. "
                    f"Hence defaulting the model to {self.model_name}",
                    UserWarning,
                )
                return False
            else:
                warnings.warn(
                    f"Unsupported model: {kwds.get('model_name')}\nSupported models are:"
                    f"\n- "
                    + "\n- ".join(list(self.model_class.keys()))
                    + "\n- All sentence-transformers models\n"
                    f"Hence defaulting to {self.model_name}",
                    UserWarning,
                )
                return False
        else:
            return False



from typing import Optional
from huggingface_hub import repo_info
import traceback
import logging
import sys

# Configure logging with colored output
logger = logging.getLogger()
logger.setLevel(logging.ERROR)

# Create console handler
console_handler = logging.StreamHandler(sys.stdout)
console_handler.setLevel(logging.DEBUG)

# Mapping of critical errors
CRITICAL_EXCEPTIONS = (
    MemoryError,
    SystemExit,
    KeyboardInterrupt,
    RuntimeError,
    ImportError,
)


# ANSI color codes
class Colors:
    RESET = "\033[0m"
    RED = "\033[91m"
    YELLOW = "\033[93m"
    CYAN = "\033[96m"
    MAGENTA = "\033[95m"
    GRAY = "\033[90m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


# Custom formatter that supports colored output
class ColoredFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: Colors.GRAY
        + "%(asctime)s - %(levelname)s - %(message)s"
        + Colors.RESET,
        logging.INFO: "%(asctime)s - %(levelname)s - %(message)s",
        logging.WARNING: Colors.YELLOW
        + "%(asctime)s - %(levelname)s - %(message)s"
        + Colors.RESET,
        logging.ERROR: Colors.RED
        + "%(asctime)s - %(levelname)s - %(message)s"
        + Colors.RESET,
        logging.CRITICAL: Colors.BOLD
        + Colors.RED
        + "%(asctime)s - %(levelname)s - %(message)s"
        + Colors.RESET,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# Apply the colored formatter to the console handler
colored_formatter = ColoredFormatter()
console_handler.setFormatter(colored_formatter)
logger.addHandler(console_handler)


def error_explainer(e: Exception) -> None:
    """
    Logs exception details with severity level and color coding based on the error type.
    Args:
        e (Exception): The exception object.
    """
    exception_type = type(e).__name__
    exception_message = str(e)

    # Extract traceback safely
    if e.__traceback__:
        exception_traceback = traceback.extract_tb(e.__traceback__)
        line_number = (
            exception_traceback[-1].lineno if exception_traceback else "Unknown"
        )
        full_traceback = "".join(traceback.format_tb(e.__traceback__))
    else:
        line_number = "Unknown"
        full_traceback = "No traceback available"

    # Colorize specific parts of the error message
    error_details = (
        f"{Colors.BOLD}Exception Type:{Colors.RESET} {Colors.MAGENTA}{exception_type}{Colors.RESET}\n"
        f"{Colors.BOLD}Exception Message:{Colors.RESET} {exception_message}\n"
        f"{Colors.BOLD}Line Number:{Colors.RESET} {Colors.CYAN}{line_number}{Colors.RESET}\n"
        f"{Colors.BOLD}Full Traceback:{Colors.RESET}\n{Colors.GRAY}{full_traceback}{Colors.RESET}"
    )

    # Decide log level based on exception type
    log_level = (
        logging.CRITICAL if isinstance(e, CRITICAL_EXCEPTIONS) else logging.ERROR
    )

    logging.log(log_level, error_details)


def hf_repo_exists(
    repo_id: str, repo_type: Optional[str] = None, token: Optional[str] = None
) -> bool:
    """Checks whether repo_id mentioned is available on huggingface
    Args:
        repo_id (str): Huggingface repo id
        repo_type (str): Type of repo
        token (str): Huggingface API token
    Returns:
        bool: True if repo exists, False otherwise
    """
    try:
        repo_info(repo_id, repo_type=repo_type, token=token)
        return True
    except Exception:
        return False
